/*
 * Copyright (c) 2010, 2011 Michael Pippig
 *
 * This file is part of PFFT.
 *
 * PFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pfft.h"
#include "ipfft.h"
#include "util.h"

static void exchange_gcells_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_along_one_dim(
    PX(gcplan) ths, int dim);
static void exchange_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim);
static void exchange_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim);
static void reduce_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim);
static void sendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset);
static void addsendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset);
static void isend_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numSend, INT sendOffset, MPI_Comm commCart1d, 
    MPI_Request *request);
static void irecv_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d, MPI_Request *request);
static void addrecv_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d);

static void create_mpi_datatype_slices(
    int rnk_n, const INT *n, INT tuple,
    int dim, INT num_slices, INT offset,
    MPI_Datatype *newtype);

/* FIXME: Generalize to arbitrary dimesions
 * Idea: canonicalize to three dims: n0 * slices * n1 */
static void add_buffer_to_slices(
    R *data, R *buffer,
    const INT *localArraySize3d, INT tupleSize,
    int dim, INT numSlices, INT offset);



void PX(exchange_gc_sendrecv)(
    PX(gcplan) ths
    )
{
  for(int dim=0; dim<ths->rnk_n; dim++)
    exchange_gcells_along_one_dim(ths, dim);
}


void PX(reduce_gc_sendrecv)(
    PX(gcplan) ths
    )
{
  for(int dim=0; dim<ths->rnk_n; dim++)
    reduce_gcells_along_one_dim(ths, dim);
}


static void exchange_gcells_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  exchange_gcells_below_along_one_dim(ths, dim);
  exchange_gcells_above_along_one_dim(ths, dim);
}


static void reduce_gcells_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  reduce_gcells_below_along_one_dim(ths, dim);
  reduce_gcells_above_along_one_dim(ths, dim);
}


static void exchange_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  sendOffset = localArrayStart;
  recvOffset = localArrayEnd;
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift++){
    numSendAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift,   ths->comms_pm[dim]);
    numRecvAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift+1, ths->comms_pm[dim]);

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    sendrecv_gcells_along_one_dim(
        ths, dim, -1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    sendOffset += numCurrentSend;
    recvOffset += numCurrentRecv;

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void exchange_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  sendOffset = localArrayEnd;
  recvOffset = localArrayStart;
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift-- ){
    numSendAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift,   ths->comms_pm[dim]);
    numRecvAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift-1, ths->comms_pm[dim]);

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    sendOffset -= numCurrentSend;
    recvOffset -= numCurrentRecv;

    sendrecv_gcells_along_one_dim(
        ths, dim, +1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void reduce_gcells_above_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd, numSendPossible, numRecvPossible;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];
  int numShifts;

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  /* corresponds to loop in exchange_gcells_above_along_one_dim 
   * with switched sender and receiver */
  numShifts = 0;
  numSendPossible = numRecvPossible = 0;
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift++){
    numSendAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift+1, ths->comms_pm[dim]);
    numRecvAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift,   ths->comms_pm[dim]);

    numSendPossible += numSendAvail;
    numRecvPossible += numRecvAvail;

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

    numShifts++;
  }

  sendOffset = localArrayEnd + ths->gc_above[dim];
  recvOffset = localArrayStart + ths->gc_above[dim];
  numSendLeftOver = numRecvLeftOver = ths->gc_above[dim];
  for(int shift = numShifts-1; shift >= 0; shift-- ){
    numSendPossible -= PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift+1, ths->comms_pm[dim]);
    numRecvPossible -= PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift,   ths->comms_pm[dim]);

    numCurrentSend = numSendLeftOver - MIN(numSendLeftOver, numSendPossible);
    numCurrentRecv = numRecvLeftOver - MIN(numRecvLeftOver, numRecvPossible);

    sendOffset -= numCurrentSend;
    recvOffset -= numCurrentRecv;

    addsendrecv_gcells_along_one_dim(
        ths, dim, +1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

  }
}


static void reduce_gcells_below_along_one_dim(
    PX(gcplan) ths, int dim
    )
{
  INT numSendLeftOver, numRecvLeftOver, sendOffset, recvOffset;
  INT numCurrentSend, numCurrentRecv, numSendAvail, numRecvAvail;
  INT localArrayStart, localArrayEnd, numSendPossible, numRecvPossible;
  INT globalArraySize = ths->n[dim], blockSize = ths->blk[dim];
  int numShifts;

  localArrayStart  = localArrayEnd = ths->gc_below[dim];
  localArrayEnd   += ths->loc_n[dim];

  /* corresponds to loop in exchange_gcells_above_along_one_dim 
   * with switched sender and receiver */
  numShifts = 0;
  numSendPossible = numRecvPossible = 0;
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift=0; (numSendLeftOver > 0) || (numRecvLeftOver > 0); shift--){
    numSendAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift-1,   ths->comms_pm[dim]);
    numRecvAvail = PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift, ths->comms_pm[dim]);

    numSendPossible += numSendAvail;
    numRecvPossible += numRecvAvail;

    numCurrentSend = MIN(numSendLeftOver, numSendAvail);
    numCurrentRecv = MIN(numRecvLeftOver, numRecvAvail);

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;

    numShifts--;
  }

  sendOffset = 0;
  recvOffset = localArrayEnd - ths->gc_below[dim];
  numSendLeftOver = numRecvLeftOver = ths->gc_below[dim];
  for(int shift = numShifts+1; shift <= 0; shift++ ){
    numSendPossible -= PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift-1,   ths->comms_pm[dim]);
    numRecvPossible -= PX(local_block_size_shifted)(
        globalArraySize, blockSize, shift, ths->comms_pm[dim]);

    numCurrentSend = numSendLeftOver - MIN(numSendLeftOver, numSendPossible);
    numCurrentRecv = numRecvLeftOver - MIN(numRecvLeftOver, numRecvPossible);

    addsendrecv_gcells_along_one_dim(
        ths, dim, -1, numCurrentSend, sendOffset, numCurrentRecv, recvOffset);

    sendOffset += numCurrentSend;
    recvOffset += numCurrentRecv;

    numSendLeftOver -= numCurrentSend;
    numRecvLeftOver -= numCurrentRecv;
  }
}


static void sendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset
    )
{
  MPI_Request mpi_req[2];

  irecv_slices(ths->data, ths->rnk_n, ths->ngc, dim, dir,
      ths->tuple, numRecv, recvOffset, ths->comms_pm[dim], &mpi_req[1]);
  isend_slices(ths->data, ths->rnk_n, ths->ngc, dim, dir,
      ths->tuple, numSend, sendOffset, ths->comms_pm[dim], &mpi_req[0]);
  MPI_Waitall(2, mpi_req, MPI_STATUSES_IGNORE);
}


static void addsendrecv_gcells_along_one_dim(
    PX(gcplan) ths, int dim, int dir,
    INT numSend, INT sendOffset, INT numRecv, INT recvOffset
    )
{
  MPI_Request request;

  isend_slices(ths->data, ths->rnk_n, ths->ngc, dim, dir,
      ths->tuple, numSend, sendOffset, ths->comms_pm[dim], &request);
  addrecv_slices(ths->data, ths->rnk_n, ths->ngc, dim, dir,
      ths->tuple, numRecv, recvOffset, ths->comms_pm[dim]);
  MPI_Wait(&request, MPI_STATUS_IGNORE);
}



static void isend_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numSend, INT sendOffset, MPI_Comm commCart1d, 
    MPI_Request *request
    )
{
  int from, to, disp;
  INT numSendTotal=tupleSize;
  MPI_Datatype sendtype;

  for(int t=0; t<rnk_n; t++)
    numSendTotal *= (dim==t) ? numSend : localArraySize[t];
  
  /* Return immediately if nothing needs to be send */
  if(numSendTotal == 0){
    *request = MPI_REQUEST_NULL;
    return;
  }

  create_mpi_datatype_slices(rnk_n, localArraySize, tupleSize, dim, numSend, sendOffset,
      &sendtype);
  MPI_Type_commit(&sendtype);

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, disp, &from, &to);

  MPI_Isend(data, 1, sendtype, to, 7, commCart1d, request);

  MPI_Type_free(&sendtype);
}


static void irecv_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d, MPI_Request *request
    )
{
  int from, to, disp;
  INT numRecvTotal=tupleSize;
  MPI_Datatype recvtype;
  
  for(int t=0; t<rnk_n; t++)
    numRecvTotal *= (dim==t) ? numRecv : localArraySize[t];
  
  /* Return immediately if nothing needs to be received */
  if(numRecvTotal == 0){
    *request = MPI_REQUEST_NULL;
    return;
  }

  create_mpi_datatype_slices(rnk_n, localArraySize, tupleSize, dim, numRecv, recvOffset,
      &recvtype);
  MPI_Type_commit(&recvtype);

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, -disp, &to, &from);

  MPI_Irecv(data, 1, recvtype, from, 7, commCart1d, request);

  MPI_Type_free(&recvtype);
}



static void addrecv_slices(
    R *data, int rnk_n, const INT *localArraySize, int dim, int dir, INT tupleSize,
    INT numRecv, INT recvOffset, MPI_Comm commCart1d
    )
{
  int from, to, disp;
  INT numRecvTotal=tupleSize;
  R *buffer;

  for(int t=0; t<rnk_n; t++)
    numRecvTotal *= (dim==t) ? numRecv : localArraySize[t];

  /* Return immediately if nothing needs to be received */
  if(numRecvTotal == 0)
    return;

  disp = (dir<0) ? -1 : +1;
  MPI_Cart_shift(commCart1d, 0, -disp, &to, &from);

  buffer = (R*) malloc(sizeof(R) * (size_t) numRecvTotal);
  MPI_Recv(buffer, (int) numRecvTotal, PFFT_MPI_REAL_TYPE, from, 7, commCart1d, MPI_STATUS_IGNORE);

  add_buffer_to_slices(data, buffer, localArraySize, tupleSize, dim, numRecv, recvOffset);
  free(buffer);
}


static void create_mpi_datatype_slices(
    int rnk_n, const INT *n, INT tuple,
    int dim, INT num_slices, INT offset,
    MPI_Datatype *newtype
    )
{
  int t;
  int *sizes, *subsizes, *starts;
  MPI_Datatype oldtype=PFFT_MPI_REAL_TYPE;

  sizes = PX(malloc_int)(rnk_n+1);
  subsizes = PX(malloc_int)(rnk_n+1);
  starts = PX(malloc_int)(rnk_n+1);

  for(t=0; t<rnk_n; t++){
    sizes[t]    = (int) ( n[t] );
    subsizes[t] = (int) ( (t==dim) ? num_slices : n[t] );
    starts[t]   = (int) ( (t==dim) ? offset    : 0 );
  }
  sizes[t] = subsizes[t] = (int) (tuple);
  starts[t]= 0;

  /* FIXME: array_of_sizes, array_of_subsizes, array_of_starts are integer */
  /* TODO:  implement MPI_Type_create_subarray with INT similar MPI_Datatypes*/
  MPI_Type_create_subarray(rnk_n+1, sizes, subsizes, starts, MPI_ORDER_C, oldtype, newtype);

  free(sizes); free(subsizes); free(starts);
}



static void add_buffer_to_slices(
    R *data, R *buffer,
    const INT *localArraySize3d, INT tupleSize,
    int dim, INT numSlices, INT offset
    )
{
  const int d=3;
  INT n[d], nSub[d], nStart[d];
  INT k0, k1, k2, h, ind;

  for(int t=0; t<3; t++){
    n[t]      = localArraySize3d[t];
    nSub[t]   = (t==dim) ? numSlices : localArraySize3d[t];
    nStart[t] = (t==dim) ? offset    : 0;
  }

  ind = 0;
  for(k0 = nStart[0]; k0 < nStart[0] + nSub[0]; k0++){
    for(k1 = nStart[1]; k1 < nStart[1] + nSub[1]; k1++){
      for(k2 = nStart[2]; k2 < nStart[2] + nSub[2]; k2++){
        for(h = 0; h < tupleSize; h++){
          data[h+tupleSize*(k2+n[2]*(k1+n[1]*k0))] += buffer[ind++];
        }
      }
    }
  }
}




