/*
 * Copyright (c) 2010, 2011 Michael Pippig
 *
 * This file is part of PFFT.
 *
 * PFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pfft.h"
#include "ipfft.h"
#include "util.h"

static sertrafo_plan plan_sertrafo_p(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in0, R *out0, R *in1, R *out1,
    int sign, const X(r2r_kind) *kinds,
    unsigned trafo_flag, unsigned transp_flag,
    unsigned fftw_flags,
    double *time);
static sertrafo_plan plan_sertrafo_pt(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in0, R *out0, R *in1, R *out1, int sign,
    const X(r2r_kind) *kinds, unsigned trafo_flag,
    unsigned transp_flag0, unsigned transp_flag1,
    unsigned fftw_flags, 
    double *time);
static sertrafo_plan plan_remap_only(
    INT nb, int rnk, const INT *n, INT howmany, R *in, R *out,
    unsigned trafo_flag, unsigned transp_flag, unsigned fftw_flags);
static X(plan) plan_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in, R *out, int sign, const X(r2r_kind) *kinds,
    unsigned trafo_flag, unsigned transp_flag, unsigned fftw_flags,
    double *time);
static X(plan) plan_remap(
    INT nb, int rnk, const INT *n, INT howmany, R *in, R *out,
    unsigned trafo_flag, unsigned transp_flag,
    unsigned fftw_flags, unsigned ioarray_flag,
    double *time);
static void malloc_and_fill_dims_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    unsigned trafo_flag, unsigned transp_flag,
    int *dims_rnk_ptr, X(iodim64) **dims_ptr,
    int *howmany_rnk_ptr, X(iodim64) **howmany_dims_ptr);
static void malloc_and_fill_dims_remap(
    INT nb, int rnk, const INT *n, INT howmany, 
    unsigned trafo_flag, unsigned transp_flag, unsigned ioarray_flag,
    int *howmany_rnk_ptr, X(iodim64) **howmany_dims_ptr);
static void malloc_and_calculate_strides_remap(
    INT nb, int rnk, const INT *pn, INT howmany,
    unsigned transp_flag,
    INT **sz_ptr, INT **is_ptr, INT **os_ptr);
static void malloc_and_calculate_strides_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    unsigned trafo_flag, unsigned transp_flag,
    INT **sz_ptr, INT **is_ptr, INT **os_ptr);
static void exchange_INT(
    INT *a, INT *b);
static double measure_time(
    X(plan) plan);
static sertrafo_plan sertrafo_mkplan(
    void);
static int needs_transpose(
    unsigned transp_flag);


/* input array:
 *   nb x n0 x n1 x ... x nr x howmany (default, PFFT_TRANSPOSE_NONE)
 *   n0 x nb x n1 x ... x nr x howmany (PFFT_TRANSPOSED_IN)
 * transformed output array:
 *   nb x N0 x N1 x ... x Nr x howmany (default, PFFT_TRANSPOSE_NONE)
 *   N0 x nb x N1 x ... x Nr x howmany (PFFT_TRANSPOSED_OUT)
 * !!! first dimension 'nb' will not be transformed !!!
 * - trafo can be c2c, r2c, c2r, r2r
 * - special attention for r2c, c2r padding
 * - r2r trafo needs r2r_kind for every dimension
 *   (different boundary conditions possible)
 * - sign only needed for c2c, otherwise sign==0
 * - kind only needed for r2r, otherwise kind==NULL
 *
 * If trafo_flag 'PFFTI_TRAFO_SKIP' is set, we execute only the
 * local transpose on the input array, but ommit the transform.
 * We use the transform type (C2C, R2C, C2R, R2R) to determine
 * the data type of the input array (and the correct strides).
 * 
 * If trafo_flag 'PFFTI_TRAFO_SKIP' AND transp_flag 'PFFT_TRANSPOSE_NONE'
 * are set, we copy the input for out-of-place plans. */

/* TODO: Calculate FFTs with stride 1 on default for both transposed in
 *       and transposed out. 
 *  T_IN: do remap first, FFT second
 *  T_OUT: do FFT first, remap second */

/* TODO: implement io_flag = {PFFT_BUFFERED_INPLACE, PFFT_PRESERVE_INPUT} */


INT PX(local_size_sertrafo)(
    INT nb, int rnk, const INT *n, INT howmany,
    unsigned trafo_flag
    )
{
  INT *pn, mem;
  
  pn = PX(malloc_INT)(rnk);

  PX(physical_dft_size)(rnk, n, trafo_flag,
      pn);

  mem = nb * PX(prod_INT)(rnk, pn) * howmany;

  free(pn);

  return mem;
}


sertrafo_plan PX(plan_sertrafo)(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in, R *out, int sign, const X(r2r_kind) *kinds,
    unsigned trafo_flag, unsigned transp_flag, unsigned io_flag,
    unsigned fftw_flags
    )
{
  /* FFTW sometimes produces very slow plans for 1dFFT with different is and os.
   * Therefore we replace the FFTW 1dFFT by two FFTW plans, one 1dFFT and one remap.
   * We can choose which one of these plans comes first, which one works out-of-place
   * and which one has different is and os. This results in at most 8 different
   * plans which we compare by time. */

  double time0, time1;
  sertrafo_plan ths0=NULL, ths1=NULL;
  
  if(trafo_flag & PFFTI_TRAFO_SKIP)
    return plan_remap_only(
        nb, rnk, n, howmany, in, out,
        trafo_flag, transp_flag, fftw_flags);

  /* try all allowed sequences of in-place and out-of-place plans */
  if(in == out) {
    if( io_flag & PFFT_BUFFERED_INPLACE ) {
      ths0 = plan_sertrafo_p(
          nb, rnk, n, howmany, in, out, out, in, sign, kinds,
          trafo_flag, transp_flag, fftw_flags,
          &time0);
      if( ~fftw_flags & FFTW_ESTIMATE ){
        ths1 = plan_sertrafo_p(
            nb, rnk, n, howmany, in, in, in, in, sign, kinds,
            trafo_flag, transp_flag, fftw_flags,
            &time1);
      } else {
        time1 = time0;
      }
    } else { /* normal inplace */
      ths0 = plan_sertrafo_p(
          nb, rnk, n, howmany, in, in, in, in, sign, kinds,
          trafo_flag, transp_flag, fftw_flags,
          &time0);
      time1 = time0;
    }
  } else {
    ths0 = plan_sertrafo_p(
        nb, rnk, n, howmany, in, out, out, out, sign, kinds,
        trafo_flag, transp_flag, fftw_flags,
        &time0);
    if( (~io_flag & PFFT_PRESERVE_INPUT) && (~fftw_flags & FFTW_ESTIMATE) ){
      ths1 = plan_sertrafo_p(
          nb, rnk, n, howmany, in, in, in, out, sign, kinds,
          trafo_flag, transp_flag, fftw_flags,
          &time1);
    } else
      time1 = time0;
  }
 
  /* choose best plan, delete the other one */
  if(time1 < time0){
    PX(sertrafo_rmplan)(ths0);
    return ths1;
  } else{
    PX(sertrafo_rmplan)(ths1);
    return ths0;
  }
}


/* only perform remap */
static sertrafo_plan plan_remap_only(
    INT nb, int rnk, const INT *n, INT howmany, R *in, R *out,
    unsigned trafo_flag, unsigned transp_flag, unsigned fftw_flags
    )
{
  double time_dummy;
  sertrafo_plan ths=NULL;
  
  ths = sertrafo_mkplan();
  ths->plan[0] = plan_remap(
      nb, rnk, n, howmany, in, out, trafo_flag, transp_flag, fftw_flags, PFFTI_ARRAY_INPUT,
      &time_dummy);
  ths->plan[1] = NULL;

  return ths;
}


static sertrafo_plan plan_sertrafo_p(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in0, R *out0, R *in1, R *out1,
    int sign, const X(r2r_kind) *kinds,
    unsigned trafo_flag, unsigned transp_flag,
    unsigned fftw_flags,
    double *time
    )
{
  unsigned transp_flag0, transp_flag1;
  double time1;
  sertrafo_plan ths0=NULL, ths1=NULL;
 
  /* Important: R2C trafo can not combine trafo and transposition.
   * Default (FFTW_ESTIMATE is set):
   * Calculate trafo with first and transpose with second plan. */ 
  
  /* transpose during 2nd plan */
  transp_flag0 = (transp_flag & PFFT_TRANSPOSED_IN) ?
    PFFT_TRANSPOSED_IN | PFFT_TRANSPOSED_OUT : PFFT_TRANSPOSED_NONE;
  transp_flag1 = transp_flag;
  ths0 = plan_sertrafo_pt(
      nb, rnk, n, howmany, in0, out0, in1, out1, sign, kinds,
      trafo_flag, transp_flag0, transp_flag1, fftw_flags,
      time);
  
  /* try this plan only for altered strides */
  if( (~fftw_flags & FFTW_ESTIMATE) && needs_transpose(transp_flag) ){
    /* transpose during 1st plan */
      transp_flag0 = transp_flag;
      transp_flag1 = (transp_flag & PFFT_TRANSPOSED_OUT) ?
        PFFT_TRANSPOSED_IN | PFFT_TRANSPOSED_OUT : PFFT_TRANSPOSED_NONE;
    ths1 = plan_sertrafo_pt(
        nb, rnk, n, howmany, in0, out0, in1, out1, sign, kinds,
        trafo_flag, transp_flag0, transp_flag1, fftw_flags,
        &time1);
  }
  else
    time1 = *time;
  
  /* choose best plan, delete the other one */
  if(time1 < *time){
    *time = time1;
    PX(sertrafo_rmplan)(ths0);
    return ths1;
  } else {
    PX(sertrafo_rmplan)(ths1);
    return ths0;
  }
}

static int needs_transpose(
    unsigned transp_flag
    )
{
  /* no transposition flag set */
  if(transp_flag == PFFT_TRANSPOSED_NONE)
    return 0;
  
  /* no transposition needed, if both flags are set */
  if(transp_flag & PFFT_TRANSPOSED_IN)
    if(transp_flag & PFFT_TRANSPOSED_OUT)
      return 0;
  
  /* need to transpose, if only one flag is set */
  return 1;
}


/* decide if FFT or remap is executed first */
static sertrafo_plan plan_sertrafo_pt(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in0, R *out0, R *in1, R *out1, int sign,
    const X(r2r_kind) *kinds, unsigned trafo_flag,
    unsigned transp_flag0, unsigned transp_flag1,
    unsigned fftw_flags, 
    double *time
    ) 
{
  double t0, t1, time0=1e10, time1=1e10;
  sertrafo_plan ths0=NULL, ths1=NULL;
  
  ths0 = sertrafo_mkplan();
     
  /* perform FFT first, remap second */
  ths0->plan[0] = plan_trafo(
      nb, rnk, n, howmany, in0, out0, sign, kinds, trafo_flag, transp_flag0, fftw_flags,
      &t0);
  ths0->plan[1] = plan_remap(
      nb, rnk, n, howmany, in1, out1, trafo_flag, transp_flag1, fftw_flags, PFFTI_ARRAY_OUTPUT,
      &t1);
  time0 = t0 + t1;
  
  /* this plan differs for altered strides or out-of-place */
  ths1 = sertrafo_mkplan();
  if(~fftw_flags & FFTW_ESTIMATE){
    if( (in0 != out1)
        || needs_transpose(transp_flag0)
        || needs_transpose(transp_flag1)
        )
    {
      /* perform remap first, FFT second */
      ths1->plan[0] = plan_remap(
          nb, rnk, n, howmany, in0, out0, trafo_flag, transp_flag0, fftw_flags, PFFTI_ARRAY_INPUT,
          &t0);
      ths1->plan[1] = plan_trafo(
          nb, rnk, n, howmany, in1, out1, sign, kinds, trafo_flag, transp_flag1, fftw_flags,
          &t1);
      time1 = t0 + t1;
    }
  }

  if(time1 < time0){
    *time = time1;
    PX(sertrafo_rmplan)(ths0);
    return ths1;
  } else {
    *time = time0;
    PX(sertrafo_rmplan)(ths1);
    return ths0;
  }
}


static X(plan) plan_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    R *in, R *out, int sign, const X(r2r_kind) *kinds,
    unsigned trafo_flag, unsigned transp_flag, unsigned fftw_flags,
    double *time
    )
{
  int dims_rnk, howmany_rnk;
  X(iodim64) *dims, *howmany_dims;
  X(plan) ths;
 
  /* R2C can not combine trafo and transposition */
  if( (trafo_flag & PFFTI_TRAFO_RDFT) && needs_transpose(transp_flag) ){
    *time = 1e10;
    return NULL;
  }

  /* R2R can not combine trafo and transposition */
  if( (trafo_flag & PFFTI_TRAFO_R2R) && needs_transpose(transp_flag) ){
    *time = 1e10;
    return NULL;
  }
  
  malloc_and_fill_dims_trafo(
    nb, rnk, n, howmany, trafo_flag, transp_flag,
    &dims_rnk, &dims, &howmany_rnk, &howmany_dims);

  /* choose appropriate fftw planner for trafo */
  if(trafo_flag & PFFTI_TRAFO_R2C){
    ths = X(plan_guru64_dft_r2c)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
        in, (C*) out, fftw_flags);
  } else if(trafo_flag & PFFTI_TRAFO_C2R){
    ths = X(plan_guru64_dft_c2r)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
        (C*) in, out, fftw_flags);
  } else if(trafo_flag & PFFTI_TRAFO_R2R){
    ths = X(plan_guru64_r2r)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
	in, out, kinds, fftw_flags);
  } else {
    ths = X(plan_guru64_dft)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
        (C*) in, (C*) out, sign, fftw_flags);
  }

  free(dims); free(howmany_dims);
  
  *time = measure_time(ths);
  return ths;
}

static void malloc_and_fill_dims_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    unsigned trafo_flag, unsigned transp_flag,
    int *dims_rnk_ptr, X(iodim64) **dims_ptr,
    int *howmany_rnk_ptr, X(iodim64) **howmany_dims_ptr
    )
{
  int dims_rnk, howmany_rnk;
  INT *sz, *is, *os;
  X(iodim64) *dims, *howmany_dims;
  
  malloc_and_calculate_strides_trafo(
    nb, rnk, n, howmany, trafo_flag, transp_flag,
    &sz, &is, &os);
    
  dims_rnk = rnk;
  dims = (X(iodim64)*) malloc(sizeof(X(iodim64)) * (size_t) dims_rnk);

  for(int t=0; t<rnk; t++){
    dims[t].n  = sz[t+1];
    dims[t].is = is[t+1];
    dims[t].os = os[t+1];
  }

  howmany_rnk = 2;
  howmany_dims = (X(iodim64)*) malloc(sizeof(X(iodim64)) * (size_t) howmany_rnk);

  /* one loop for nb */
  howmany_dims[0].n  = sz[0];
  howmany_dims[0].is = is[0];
  howmany_dims[0].os = os[0];

  /* one loop for howmany */
  howmany_dims[1].n  = sz[rnk+1];
  howmany_dims[1].is = is[rnk+1];
  howmany_dims[1].os = os[rnk+1];

  *dims_ptr = dims;
  *howmany_dims_ptr = howmany_dims;
  *dims_rnk_ptr = dims_rnk;
  *howmany_rnk_ptr = howmany_rnk;

  free(sz); free(is); free(os);
}

/* We use a r2r plan for r2c transposition. For the input
 * we double the physical array size and for the output we
 * double 'howmany'. Analogously done for c2r. */
static X(plan) plan_remap(
    INT nb, int rnk, const INT *n, INT howmany, R *in, R *out,
    unsigned trafo_flag, unsigned transp_flag,
    unsigned fftw_flags, unsigned ioarray_flag,
    double *time
    )
{
  /* wrap FFTWs guru64 FFT planing, only reorder without FFT */
  int sign=1, howmany_rnk, dims_rnk=0;
  X(r2r_kind) *kinds=NULL;
  X(iodim64) *howmany_dims, *dims=NULL;
  X(plan) ths;
    
  malloc_and_fill_dims_remap(
      nb, rnk, n, howmany, trafo_flag, transp_flag, ioarray_flag,
      &howmany_rnk, &howmany_dims);

  /* choose appropriate fftw planner for remap */
  if(trafo_flag & PFFTI_TRAFO_C2C)
    ths = X(plan_guru64_dft)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
        (C*) in, (C*) out, sign, fftw_flags);
  else
    ths = X(plan_guru64_r2r)(
        dims_rnk, dims, howmany_rnk, howmany_dims,
	in, out, kinds, fftw_flags);
    
  free(howmany_dims);

  *time = measure_time(ths);
  return ths;
}


static void malloc_and_fill_dims_remap(
    INT nb, int rnk, const INT *n, INT howmany, 
    unsigned trafo_flag, unsigned transp_flag, unsigned ioarray_flag,
    int *howmany_rnk_ptr, X(iodim64) **howmany_dims_ptr
    )
{
  int howmany_rnk;
  INT *sz, *is, *os, *pn;
  X(iodim64) *howmany_dims;

  pn = PX(malloc_INT)(rnk);
  PX(physical_dft_size)(rnk, n, trafo_flag, pn);

  if( trafo_flag & PFFTI_TRAFO_R2C ){
    if( ioarray_flag & PFFTI_ARRAY_OUTPUT )
      howmany *= 2;
    else
      pn[rnk-1] *= 2;
  }
  
  if( trafo_flag & PFFTI_TRAFO_C2R ){
    if( ioarray_flag & PFFTI_ARRAY_INPUT )
      howmany *= 2;
    else
      pn[rnk-1] *= 2;
  }

  malloc_and_calculate_strides_remap(
    nb, rnk, pn, howmany, transp_flag,
    &sz, &is, &os);
 
  howmany_rnk = rnk+2;
  howmany_dims = (X(iodim64)*) malloc(sizeof(X(iodim64)) * (size_t) howmany_rnk);

  for(int t=0; t<rnk+2; t++){
    howmany_dims[t].n  = sz[t];
    howmany_dims[t].is = is[t];
    howmany_dims[t].os = os[t];
  }

  *howmany_dims_ptr = howmany_dims;
  *howmany_rnk_ptr = howmany_rnk;

  free(sz); free(is); free(os); free(pn);
}

/* set array size 'sz' for array of size
 * nb x n[0] x n[1] x ... x n[rnk-1] x howmany,
 * set input strides 'is' corresponing to array of size
 * nb x n[0] x n[1] x ... x n[rnk-1] x howmany (default) or
 * n[0] x nb x n[1] x ... x n[rnk-1] x howmany (TRANSPOSED_IN),
 * set output strides 'os' corresponing to array of size
 * nb x n[0] x n[1] x ... x n[rnk-1] x howmany (default) or
 * n[0] x nb x n[1] x ... x n[rnk-1] x howmany (TRANSPOSED_OUT),
 */
static void malloc_and_calculate_strides_remap(
    INT nb, int rnk, const INT *pn, INT howmany,
    unsigned transp_flag,
    INT **sz_ptr, INT **is_ptr, INT **os_ptr
    )
{ 
  INT *sz, *is, *os, *pni, *pno;

  /* allocate return arrays */
  sz = PX(malloc_INT)(rnk+2);
  is = PX(malloc_INT)(rnk+2);
  os = PX(malloc_INT)(rnk+2);

  /* we need the physical array sizes to compute strides */
  pni = PX(malloc_INT)(rnk+2);
  pno = PX(malloc_INT)(rnk+2);
  
  /* set FFTW array size */
  sz[0] = nb;
  for(int t=0; t<rnk; t++)
    sz[t+1] = pn[t];
  sz[rnk+1] = howmany;

  /* set physical array size */
  for(int t=0; t<rnk+2; t++)
    pni[t] = pno[t] = sz[t];

  /* permute first two dims of array size */
  if(transp_flag & PFFT_TRANSPOSED_IN )
    exchange_INT(&pni[0], &pni[1]);
  if(transp_flag & PFFT_TRANSPOSED_OUT )
    exchange_INT(&pno[0], &pno[1]);

  /* calculate strides */
  is[rnk+1] = os[rnk+1] = 1;
  for(int t=rnk; t>=0; t--){
    is[t] = is[t+1] * pni[t+1];
    os[t] = os[t+1] * pno[t+1];
  }
  
  /* permute first two dims of strides */
  if(transp_flag & PFFT_TRANSPOSED_IN )
    exchange_INT(&is[0], &is[1]);
  if(transp_flag & PFFT_TRANSPOSED_OUT )
    exchange_INT(&os[0], &os[1]);

  free(pni); free(pno);

  *sz_ptr = sz;
  *is_ptr = is;
  *os_ptr = os;
}


static void malloc_and_calculate_strides_trafo(
    INT nb, int rnk, const INT *n, INT howmany,
    unsigned trafo_flag, unsigned transp_flag,
    INT **sz_ptr, INT **is_ptr, INT **os_ptr
    )
{ 
  INT *sz, *is, *os, *pni, *pno;

  /* allocate return arrays */
  sz = PX(malloc_INT)(rnk+2);
  is = PX(malloc_INT)(rnk+2);
  os = PX(malloc_INT)(rnk+2);

  /* we need the physical array sizes to compute strides */
  pni = PX(malloc_INT)(rnk+2);
  pno = PX(malloc_INT)(rnk+2);
  
  /* set FFTW array size */
  sz[0] = nb;
  for(int t=0; t<rnk; t++)
    sz[t+1] = n[t];
  sz[rnk+1] = howmany;

  /* set physical array size */
  for(int t=0; t<rnk+2; t++)
    pni[t] = pno[t] = sz[t];

  /* trafo of last dim maybe r2c or c2r */
  /* compute physical size of R2C for real input and complex output */
  if(trafo_flag & PFFTI_TRAFO_R2C){
    pni[rnk] = 2 * PX(physical_dft_size_1d)(sz[rnk], trafo_flag);
    pno[rnk] =  PX(physical_dft_size_1d)(sz[rnk], trafo_flag);
  }

  /* compute physical size of C2R for complex input and real output */
  if(trafo_flag & PFFTI_TRAFO_C2R){
    pni[rnk] = PX(physical_dft_size_1d)(sz[rnk], trafo_flag);
    pno[rnk] = 2 * PX(physical_dft_size_1d)(sz[rnk], trafo_flag);
  }

  /* permute first two dims of array size */
  if(transp_flag & PFFT_TRANSPOSED_IN )
    exchange_INT(&pni[0], &pni[1]);
  if(transp_flag & PFFT_TRANSPOSED_OUT )
    exchange_INT(&pno[0], &pno[1]);

  /* calculate strides */
  is[rnk+1] = os[rnk+1] = 1;
  for(int t=rnk; t>=0; t--){
    is[t] = is[t+1] * pni[t+1];
    os[t] = os[t+1] * pno[t+1];
  }
  
  /* permute first two dims of strides */
  if(transp_flag & PFFT_TRANSPOSED_IN )
    exchange_INT(&is[0], &is[1]);
  if(transp_flag & PFFT_TRANSPOSED_OUT )
    exchange_INT(&os[0], &os[1]);

  free(pni); free(pno);

  *sz_ptr = sz;
  *is_ptr = is;
  *os_ptr = os;
}

static void exchange_INT(
    INT *a, INT *b
    )
{
  INT buf = *a;
  *a = *b;
  *b = buf;
}

static double measure_time(
    X(plan) plan
    )
{
  double time=0;
 
  if(plan != NULL){
    /* 1st execution is often very slow, discard it */
    X(execute)(plan);
    /* measure 2nd execution */
    time = -MPI_Wtime();
    X(execute)(plan);
    time += MPI_Wtime();
  }

  return time;
}


static sertrafo_plan sertrafo_mkplan(
    void
    )
{
  sertrafo_plan ths = (sertrafo_plan) malloc(sizeof(sertrafo_plan_s));

  /* initialize to NULL for easy checks */
  for(int t=0; t<2; t++)
    ths->plan[t]=NULL;
  
  return ths;
}


void PX(sertrafo_rmplan)(
    sertrafo_plan ths
    )
{
  /* plan was already destroyed or never initialized */
  if(ths==NULL)
    return;

  /* take care of unsuccessful FFTW planing */
  for(int t=0; t<2; t++)
    if(ths->plan[t] != NULL)
      X(destroy_plan)(ths->plan[t]);

  /* free memory */
  free(ths);
  /* ths=NULL; would be senseless, since we can not change the pointer itself */
}


void PX(execute_sertrafo)(
    sertrafo_plan ths
    )
{
  if(ths==NULL)
    return;

  /* execute all initialized serfft plans */
  for(int t=0; t<2; t++)
    if(ths->plan[t] != NULL)
      X(execute)(ths->plan[t]);
}









