/*
 * Copyright (c) 2010, 2011 Michael Pippig
 *
 * This file is part of PFFT.
 *
 * PFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pfft.h"
#include "ipfft.h"

static void get_global_transp_param(
    int step, int rnk_pm, const INT *ni, const INT *no,
    const INT *local_ni, const INT *local_no,
    INT tuple_size, const INT *iblock, const INT *oblock,
    unsigned trafo_flag,
    INT *N0, INT *N1, INT *h0, INT *h1,
    INT *hm, INT *blk0, INT *blk1);


/* MPI Transpose for real data. Use hm==2 for complex data.  */
/* Input:
 *   (N0 x h0) x (N1/P x h1) x hm  (default)
 *   (N1/P x h1) x (N0 x h0) x hm (TRANSPOSED_IN)
 * Output:
 *   (N1 x h1) x (N0/P x h0) x hm (default)
 *   (N0/P x h0) x (N1 x h1) x hm (TRANSPOSED_OUT)
 *  
 *  This is a collective routine.
 *  N0, h0, N1, h1 must be equal on all calling processes.
 *   */


void PX(get_global_transp_param)(
    int step, int rnk_pm, const INT *ni, const INT *no,
    const INT *local_ni, const INT *local_no,
    INT tuple_size, const INT *iblock, const INT *oblock,
    unsigned trafo_flag, unsigned transp_flag,
    INT *N0, INT *N1, INT *h0, INT *h1,
    INT *hm, INT *blk0, INT *blk1
    )
{
  /* TRANSPOSED_IN is planned in backward direction.
   * Therefore switch input and output parameters. */
  if(transp_flag & PFFT_TRANSPOSED_IN)
    get_global_transp_param(
        step, rnk_pm, no, ni, local_no, local_ni,
        tuple_size, oblock, iblock, trafo_flag,
	N1, N0, h1, h0, hm, blk1, blk0);
  else
    get_global_transp_param(
        step, rnk_pm, ni, no, local_ni, local_no,
        tuple_size, iblock, oblock, trafo_flag,
	N0, N1, h0, h1, hm, blk0, blk1);
}


static void get_global_transp_param(
    int step, int rnk_pm, const INT *ni, const INT *no,
    const INT *local_ni, const INT *local_no,
    INT tuple_size, const INT *iblock, const INT *oblock,
    unsigned trafo_flag,
    INT *N0, INT *N1, INT *h0, INT *h1,
    INT *hm, INT *blk0, INT *blk1
    )
{
  *N0 = no[rnk_pm     - step];
  *N1 = ni[rnk_pm - 1 - step];

  *h0 = 1;
  for(int t=0; t<step; t++)
    *h0 *= local_no[rnk_pm - t];
  for(int t=0; t<rnk_pm-1-step; t++)
    *h0 *= local_ni[t];
  *h1 = tuple_size;
  
  /* double hm for complex data */
  *hm = (~trafo_flag & PFFTI_TRAFO_R2R) ? 2 : 1;

  *blk0 = oblock[rnk_pm - 1 - step];
  *blk1 = iblock[rnk_pm - 1 - step];
}


INT PX(local_size_global_transp)(
    INT N0, INT N1, INT h0, INT h1, INT hm, INT blk0, INT blk1,
    MPI_Comm comm 
    )
{
  INT N[2], blk[2];
  INT dummy0, dummy1, dummy2, dummy3;
  INT mem = 1; /* never return 0 */  

  N[0] = N1 * h1; blk[0] = blk1 * h1;
  N[1] = N0 * h0; blk[1] = blk0 * h0;

  /* For strange distributions (e.g. 4 data points on 3 processes)
   * all processes in a row get no data. */
  if(N[0]*N[1]*h0*h1*hm==0)
    return 1;
  
  mem = XM(local_size_many_transposed)(
      2, N, hm, blk[0], blk[1], comm,
      &dummy0, &dummy1, &dummy2, &dummy3);

  return MAX(mem, 1);
}


X(plan) PX(plan_global_transp)(
    INT N0, INT N1, INT h0, INT h1, INT hm, INT blk0, INT blk1,
    MPI_Comm comm, R *in, R *out,
    unsigned transp_flag, unsigned fftw_flags
    )
{
  INT N[2], blk[2];

  N[0] = N1 * h1; blk[0] = blk1 * h1;
  N[1] = N0 * h0; blk[1] = blk0 * h0;

  /* For strange distributions (e.g. 4 data points on 3 processes)
   * all processes in a row get no data. */
  if(N[0]*N[1]*h0*h1*hm==0)
    return NULL;

  /* PFFTs and FFTWs transpose flags are complementary */
  if( ~transp_flag & PFFT_TRANSPOSED_OUT)
    fftw_flags |= FFTW_MPI_TRANSPOSED_OUT;
  if( ~transp_flag & PFFT_TRANSPOSED_IN)
    fftw_flags |= FFTW_MPI_TRANSPOSED_IN;

/******************************************************
 * Sometimes paralllel FFTW runs into a deadlock,
 * which can be avoided by calling fftw_forget_wisdom
 * FIXME: write bugreport to FFTW 
 ******************************************************/
X(forget_wisdom)();

  return XM(plan_many_transpose)(
      N[0], N[1], hm, blk[0], blk[1], in, out,
      comm, fftw_flags);
}

